/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** @file 
    @brief Basic voxel-based image registration methods.

    @warning This code is BETA and it may change significantly

    @author Ondrej Danek (xdanek2@fi.muni.cz)
    @date 2010
*/ 

#ifdef __GNUG__
    #pragma implementation "Registration.h"
#endif

#include <math.h>
#include "Registration.h"

namespace i3d
{
    namespace Registration
    {
        //////////////////////////////////////////////////////////////////////////
        //
        // SIMILARITY METRICS
        //
        //////////////////////////////////////////////////////////////////////////

        /******************************************************************************/

        template <class T, class VOXEL>
        ISimilarityMetric<T,VOXEL>::ISimilarityMetric(const Image3d<VOXEL> &ref_img, const Region &region)
            : m_ref_img(ref_img), m_region(region)
        {
            if (region.m_ofs.x + region.m_sz.x > ref_img.GetSizeX() ||
                region.m_ofs.y + region.m_sz.y > ref_img.GetSizeY() ||
                region.m_ofs.z + region.m_sz.z > ref_img.GetSizeZ() ||
                !region.m_sz.x || !region.m_sz.y || !region.m_sz.z)
            {
                throw InternalException("ISimilarityMetric: Invalid region.");
            }
        }

        /******************************************************************************/

        template <class T, class VOXEL>
        IGenericMutualInformation<T,VOXEL>::IGenericMutualInformation (const Image3d<VOXEL> &ref_img, 
            const Region &region, size_t levels)
            : ISimilarityMetric<T,VOXEL>(ref_img, region), m_levels(levels)
        {
            if (!m_levels)
            {
                throw InternalException("MutualInformation: Number of levels can not be zero.");
            }

            m_hist_im = NULL;
            m_hist_joint = NULL;

            try
            {
                m_hist_im = new size_t[m_levels];
                m_hist_joint = new size_t[m_levels * m_levels];
            }
            catch (std::bad_alloc)
            {
                if (m_hist_im != NULL)
                {
                    delete [] m_hist_im;
                }

                throw InternalException("MutualInformation: Not enough memory.");
            }
        }

        /******************************************************************************/

        template <class T, class VOXEL>
        T IGenericMutualInformation<T,VOXEL>::MarginalEntropy(const Image3d<VOXEL> &img, 
            const Vector3d<size_t> &ofs)
        {            
            // Compute histogram
            for (size_t i = 0; i < m_levels; i++)
            {
                m_hist_im[i] = 0;
            }

            size_t add_y = img.GetSizeX() - m_region.m_sz.x;
            size_t add_z = (img.GetSizeY() - m_region.m_sz.y) * img.GetSizeX();
            const VOXEL *ip = img.GetFirstVoxelAddr() + img.GetIndex(ofs);

            for (size_t z = 0; z < m_region.m_sz.z; z++)
            {
                for (size_t y = 0; y < m_region.m_sz.y; y++)
                {
                    for (size_t x = 0; x < m_region.m_sz.x; x++)
                    {
                        m_hist_im[*ip]++;
                        ip++;
                    }

                    ip += add_y;
                }

                ip += add_z;
            }

            return Entropy(m_hist_im, m_levels);
        }

        /******************************************************************************/

        template <class T, class VOXEL>
        T IGenericMutualInformation<T,VOXEL>::Compute(const Image3d<VOXEL> &img, 
            const Vector3d<int> &shift)
        {
            // Compute Hy and joint entropy            
            for (size_t i = 0; i < m_levels; i++)
            {
                m_hist_im[i] = 0;
            }

            for (size_t i = m_levels * m_levels; i-- > 0; )
            {
                m_hist_joint[i] = 0;
            }

            Vector3d<size_t> ofs(Vector3d<int>(m_region.m_ofs) - shift);
            size_t add_y = img.GetSizeX() - m_region.m_sz.x;
            size_t add_z = (img.GetSizeY() - m_region.m_sz.y) * img.GetSizeX();

            const VOXEL *ip1 = m_ref_img.GetFirstVoxelAddr() + m_ref_img.GetIndex(m_region.m_ofs);
            const VOXEL *ip2 = img.GetFirstVoxelAddr() + img.GetIndex(ofs);

            for (size_t z = 0; z < m_region.m_sz.z; z++)
            {
                for (size_t y = 0; y < m_region.m_sz.y; y++)
                {
                    for (size_t x = 0; x < m_region.m_sz.x; x++)
                    {
                        m_hist_im[*ip2]++;
                        m_hist_joint[(*ip1) * m_levels + (*ip2)]++;
                        ip1++;
                        ip2++;
                    }

                    ip1 += add_y;
                    ip2 += add_y;
                }

                ip1 += add_z;
                ip2 += add_z;
            }

            T Hy = Entropy(m_hist_im, m_levels);
            T Hxy = Entropy(m_hist_joint, m_levels * m_levels);

            return -(m_Hx + Hy - Hxy);
        }

        /******************************************************************************/

        template <class T, class VOXEL>
        T MutualInformation<T,VOXEL>::Entropy(size_t *hist, size_t levels) const
        {
            T ent = 0, p;
            T num_vox = T(m_region.m_sz.x * m_region.m_sz.y * m_region.m_sz.z);

            for (size_t i = 0; i < levels; i++)
            {
                if (hist[i] > 0)
                {
                    p = T(hist[i]) / num_vox;
				    ent -= p * log(p);
                }
            }

            return ent;
        }

        /******************************************************************************/

        // Explicit instantiations
        template class I3D_ALGO_EXPORT MutualInformation<float, GRAY8>;
        template class I3D_ALGO_EXPORT MutualInformation<float, GRAY16>;
        template class I3D_ALGO_EXPORT MutualInformation<double, GRAY8>;
        template class I3D_ALGO_EXPORT MutualInformation<double, GRAY16>;

        /******************************************************************************/

        template <class T, class VOXEL>
        T GeneralizedMutualInformation<T,VOXEL>::Entropy(size_t *hist, size_t levels) const
        {
            T ent = 0, p;
            T num_vox = T(m_region.m_sz.x * m_region.m_sz.y * m_region.m_sz.z);

            for (size_t i = 0; i < levels; i++)
            {
                p = T(hist[i]) / num_vox;
                ent += p * p;
            }

            return -log(ent);
        }

        /******************************************************************************/

        // Explicit instantiations
		template class I3D_ALGO_EXPORT IGenericMutualInformation<float, GRAY8>;
        template class I3D_ALGO_EXPORT IGenericMutualInformation<float, GRAY16>;
        template class I3D_ALGO_EXPORT IGenericMutualInformation<double, GRAY8>;
        template class I3D_ALGO_EXPORT IGenericMutualInformation<double, GRAY16>;
        template class I3D_ALGO_EXPORT GeneralizedMutualInformation<float, GRAY8>;
        template class I3D_ALGO_EXPORT GeneralizedMutualInformation<float, GRAY16>;
        template class I3D_ALGO_EXPORT GeneralizedMutualInformation<double, GRAY8>;
        template class I3D_ALGO_EXPORT GeneralizedMutualInformation<double, GRAY16>;

        /******************************************************************************/

        template <class T, class VOXEL>
        NormalizedCorrelation<T,VOXEL>::NormalizedCorrelation(const Image3d<VOXEL> &ref_img, 
            const Region &region)
            : ISimilarityMetric<T,VOXEL>(ref_img, region)
        {
            m_ref_mean = GetMeanValue(ref_img, region.m_ofs);
        }

        /******************************************************************************/

        template <class T, class VOXEL>
        T NormalizedCorrelation<T,VOXEL>::GetMeanValue(const Image3d<VOXEL> &img, 
            const Vector3d<size_t> &ofs)
        {            
            T isum = 0;
            size_t add_y = img.GetSizeX() - m_region.m_sz.x;
            size_t add_z = (img.GetSizeY() - m_region.m_sz.y) * img.GetSizeX();
            const VOXEL *ip = img.GetFirstVoxelAddr() + img.GetIndex(ofs);

            for (size_t z = 0; z < m_region.m_sz.z; z++)
            {
                for (size_t y = 0; y < m_region.m_sz.y; y++)
                {
                    for (size_t x = 0; x < m_region.m_sz.x; x++)
                    {
                        isum += T(*ip);
                        ip++;
                    }

                    ip += add_y;
                }

                ip += add_z;
            }

            return isum / T(m_region.m_sz.x * m_region.m_sz.y * m_region.m_sz.z);
        }

        /******************************************************************************/

        template <class T, class VOXEL>
        T NormalizedCorrelation<T,VOXEL>::Compute(const Image3d<VOXEL> &img, 
            const Vector3d<int> &shift)
        {
            Vector3d<size_t> ofs(Vector3d<int>(m_region.m_ofs) - shift);

            T img_mean = GetMeanValue(img, ofs), nom = 0, c1, c2, var1 = 0, var2 = 0;

            size_t add_y = img.GetSizeX() - m_region.m_sz.x;
            size_t add_z = (img.GetSizeY() - m_region.m_sz.y) * img.GetSizeX();

            const VOXEL *ip1 = m_ref_img.GetFirstVoxelAddr() + m_ref_img.GetIndex(m_region.m_ofs);
            const VOXEL *ip2 = img.GetFirstVoxelAddr() + img.GetIndex(ofs);

            for (size_t z = 0; z < m_region.m_sz.z; z++)
            {
                for (size_t y = 0; y < m_region.m_sz.y; y++)
                {
                    for (size_t x = 0; x < m_region.m_sz.x; x++)
                    {
                        c1 = T(*ip1) - m_ref_mean;
                        c2 = T(*ip2) - img_mean;
                        nom += c1 * c2;
                        var1 += c1 * c1;
                        var2 += c2 * c2;
                        ip1++;
                        ip2++;
                    }

                    ip1 += add_y;
                    ip2 += add_y;
                }

                ip1 += add_z;
                ip2 += add_z;
            }

            return -(nom / sqrt(var1 * var2));
        }

        /******************************************************************************/

        // Explicit instantiations
        template class I3D_ALGO_EXPORT NormalizedCorrelation<float, GRAY8>;
        template class I3D_ALGO_EXPORT NormalizedCorrelation<float, GRAY16>;
        template class I3D_ALGO_EXPORT NormalizedCorrelation<double, GRAY8>;
        template class I3D_ALGO_EXPORT NormalizedCorrelation<double, GRAY16>;

        /******************************************************************************/

        //////////////////////////////////////////////////////////////////////////
        //
        // MINIMIZATION ALGORITHMS
        //
        //////////////////////////////////////////////////////////////////////////

        /******************************************************************************/

        template <class T, class VOXEL>
        void BruteForceSearch<T,VOXEL>::Minimize(const Image3d<VOXEL> &img, 
            ISimilarityMetric<T,VOXEL> &sim, const Transformation<T> &tmin, 
            const Transformation<T> &tmax, Transformation<T> &opt)
        {
            T best = std::numeric_limits<T>::max();

            for (T z = tmin.m_shift.z; z <= tmax.m_shift.z; z += m_step.m_shift.z)
            {
                for (T y = tmin.m_shift.y; y <= tmax.m_shift.y; y += m_step.m_shift.y)
                {
                    for (T x = tmin.m_shift.x; x <= tmax.m_shift.x; x += m_step.m_shift.x)
                    {
                        T sf = sim.Compute(img, Vector3d<int>(int(x), int(y), int(z)));

                        if (sf < best)
                        {
                            best = sf;
                            opt.m_shift = Vector3d<T> (x,y,z);
                        }
                    }
                }
            }
        }

        /******************************************************************************/

        // Explicit instanitations
        template class I3D_ALGO_EXPORT BruteForceSearch<float, GRAY8>;
        template class I3D_ALGO_EXPORT BruteForceSearch<float, GRAY16>;
        template class I3D_ALGO_EXPORT BruteForceSearch<double, GRAY8>;
        template class I3D_ALGO_EXPORT BruteForceSearch<double, GRAY16>;

        /******************************************************************************/

        template <class T, class VOXEL>
        void GradientDescent<T,VOXEL>::Minimize(const Image3d<VOXEL> &img, 
            ISimilarityMetric<T,VOXEL> &sim, const Transformation<T> &tmin, 
            const Transformation<T> &tmax, Transformation<T> &opt)
        {
            bool improvement = true;
            T best, sf;
            Transformation<T> bt;
            opt.m_shift = (tmin.m_shift + tmax.m_shift) / T(2);

            best = sim.Compute(img, Vector3d<int>(opt.m_shift));;

            while (improvement)
            {
                improvement = false;                

                int x = int(opt.m_shift.x);
                int y = int(opt.m_shift.y);
                int z = int(opt.m_shift.z);

                if (opt.m_shift.x - 1 >= tmin.m_shift.x)
                {
                    sf = sim.Compute(img, Vector3d<int>(x-1,y,z));
                    if (sf < best)
                    {
                        best = sf;
                        bt = opt;
                        bt.m_shift.x--;
                        improvement = true;
                    }
                }

                if (opt.m_shift.x + 1 <= tmax.m_shift.x)
                {
                    sf = sim.Compute(img, Vector3d<int>(x+1,y,z));
                    if (sf < best)
                    {
                        best = sf;
                        bt = opt;
                        bt.m_shift.x++;
                        improvement = true;
                    }
                }

                if (opt.m_shift.y - 1 >= tmin.m_shift.y)
                {
                    sf = sim.Compute(img, Vector3d<int>(x,y-1,z));
                    if (sf < best)
                    {
                        best = sf;
                        bt = opt;
                        bt.m_shift.y--;
                        improvement = true;
                    }
                }

                if (opt.m_shift.y + 1 <= tmax.m_shift.y)
                {
                    sf = sim.Compute(img, Vector3d<int>(x,y+1,z));
                    if (sf < best)
                    {
                        best = sf;
                        bt = opt;
                        bt.m_shift.y++;
                        improvement = true;
                    }
                }

                if (opt.m_shift.z - 1 >= tmin.m_shift.z)
                {
                    sf = sim.Compute(img, Vector3d<int>(x,y,z-1));
                    if (sf < best)
                    {
                        best = sf;
                        bt = opt;
                        bt.m_shift.z--;
                        improvement = true;
                    }
                }

                if (opt.m_shift.z + 1 <= tmax.m_shift.z)
                {
                    sf = sim.Compute(img, Vector3d<int>(x,y,z+1));
                    if (sf < best)
                    {
                        best = sf;
                        bt = opt;
                        bt.m_shift.z++;
                        improvement = true;
                    }
                }

                if (improvement)
                {
                    opt = bt;
                }
            }
        }

        /******************************************************************************/

        // Explicit instanitations
        template class I3D_ALGO_EXPORT GradientDescent<float, GRAY8>;
        template class I3D_ALGO_EXPORT GradientDescent<float, GRAY16>;
        template class I3D_ALGO_EXPORT GradientDescent<double, GRAY8>;
        template class I3D_ALGO_EXPORT GradientDescent<double, GRAY16>;

        /******************************************************************************/
    }
}

