/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** @file 
    @brief Kuhn-Munkres Algorithm

    (H. W. Kuhn, The Hungarian method for the assignment problem,
    Naval Research Logistics Quarterly, 2 (1955), pp. 83-97. )

    Based on description available from:
    http://www.math.uwo.ca/~mdawes/courses/344/kuhn-munkres.pdf

    @author Petr Matula (pem@fi.muni.cz)
    @author Ondrej Danek (xdanek2@fi.muni.cz)
    @date 2010
*/ 

#include <algorithm>
#include <stack>
#include <set>
#include <vector>
#include "basic.h"
#include "RegistrationKuhnMunkres.h"

namespace i3d
{
    namespace Registration
    {
        /******************************************************************************/

        static const double SMALL_DIFF = 1e-10;

        /******************************************************************************/

        class BipartiteGraph 
        {
        private:
            std::vector<std::set<size_t> > l;
            size_t maxr;

        public:
            typedef std::set<size_t>::const_iterator const_iterator;

            BipartiteGraph()
                : maxr(0) 
            {}            

            size_t size() const { return l.size(); }
            size_t rsize() const { return maxr; }
            void resize(size_t n) { l.resize(n); }

            void delete_all_edges() {
                for (size_t i = 0; i < l.size(); ++i)
                {
                    l[i].clear();
                }
            }

            void add_edge(size_t i, size_t j) {
                l[i].insert(j);
                maxr = std::max(maxr,j);
            }

            bool has_edge(size_t i, size_t j) const {
                return find(l[i].begin(),l[i].end(),j) != l[i].end();
            }

            const std::set<size_t> & edges(size_t i) const { return l[i]; }
        };

        /******************************************************************************/

        class Matching 
        {
        private:
            std::vector<size_t> left;
            std::vector<size_t> right;
            std::set<size_t> unmatched;

        public:
            Matching(size_t n) {
                left.resize(n);
                right.resize(n);
                for (size_t i = 0; i < n; ++i)
                {
                    unmatched.insert(i);
                    right[i] = std::numeric_limits<size_t>::max();
                    left[i] = std::numeric_limits<size_t>::max();
                }
            }

            size_t size() const { return left.size(); }

            bool add_edge(size_t i, size_t j) {
                if (left[i] == std::numeric_limits<size_t>::max() && 
                    right[j] == std::numeric_limits<size_t>::max())
                {
                    left[i] = j;
                    right[j] = i;
                    unmatched.erase(i);
                    return true;
                }
                return false;
            }

            void delete_edge_left(size_t i) {
                size_t j = left[i];
                if (j != std::numeric_limits<size_t>::max())
                {
                    right[j] = std::numeric_limits<size_t>::max();
                    left[i] = std::numeric_limits<size_t>::max();
                    unmatched.insert(i);
                }
            }

            bool is_complete() const { return unmatched.empty(); }
            size_t get_unmatched() const { return *unmatched.begin(); }
            size_t left_of(size_t i) const { return right[i]; }
            size_t unused() const { return std::numeric_limits<size_t>::max(); }

            const std::vector<size_t> &matching() const { return left; }
        };

        /******************************************************************************/

        template <class DT>
        static void EqualitySubgraph(const std::vector<DT> &left,
            const std::vector<DT> &right,
            const std::vector<DT> &weights,
            BipartiteGraph &gl)
        {
            gl.resize(left.size());
            gl.delete_all_edges();

            const DT *wp = &weights[0];

            for (size_t i = 0; i < left.size(); ++i)
            {
                for (size_t j = 0; j < right.size(); ++j, wp++)
                {
                    if (left[i] + right[j] - *wp < SMALL_DIFF)
                    {
                        gl.add_edge(i,j);
                    }
                }
            }
        }

        /******************************************************************************/

        static void Enlarge_J(size_t x, const BipartiteGraph &gl, std::set<size_t> &J)
        {
            BipartiteGraph::const_iterator bgi;

            // J = { y | (x,y), x in S, y in Y}
            for (bgi = gl.edges(x).begin(); bgi != gl.edges(x).end(); ++bgi)
                J.insert(*bgi);
        }

        /******************************************************************************/

        static void ArbitraryMatching(const BipartiteGraph &gl, Matching &m)
        {
            BipartiteGraph::const_iterator j;

            for (size_t i = 0; i < gl.size(); ++i)
                for (j = gl.edges(i).begin(); j != gl.edges(i).end(); ++j)
                    if (m.add_edge(i,*j))
                        break;
        }

        /******************************************************************************/

        static void Complement(const std::set<size_t> &T, size_t n, std::set<size_t> &Tc)
        {
            std::set<size_t>::const_iterator t;

            Tc.clear();
            t = T.begin();


            for (size_t i = 0; i < n; i++)
                if (t != T.end() && i == *t)
                    ++t;
                else
                    Tc.insert(i);
        }

        /******************************************************************************/

        template <class DT>
        static DT GetAlpha(const std::vector<DT> &left, const std::vector<DT> &right,
            const std::set<size_t> &S, const std::set<size_t> &Tc, 
            size_t n, const std::vector<DT> &weights)
        {
            std::set<size_t>::const_iterator t;
            std::set<size_t>::const_iterator s;

            DT m = std::numeric_limits<DT>::max();

            for (s = S.begin(); s != S.end(); ++s)
            {
                for (t = Tc.begin(); t != Tc.end(); ++t)
                {
                    DT tmp = left[*s] + right[*t] - weights[(*s) * n + (*t)];
                    if (m > tmp)
                        m = tmp;
                }
            }

            return m;
        }

        /******************************************************************************/

        static void ExpandMatching(Matching &m, const BipartiteGraph &gl, size_t x, size_t y)
        {
            // find an alternating path p in gl from x to y
            // p = {(u1, v1), (u2, v2), ... (un, vn)}, n is odd, is alternating
            // iff (ui, vi) is in M for even i and is not in M for odd i
            // let E be the set of edges which lie on p
            // let M be the set of edges in matching m
            // let I = (M intersect E)
            // remove edges I from m
            // add edges E \ I to m
            std::stack<size_t> node;
            std::stack<std::set<size_t>::const_iterator> last;  // Change from iterator to const_iterator
            size_t idx;
            node.push(x);
            last.push(gl.edges(x).begin());
            std::vector<bool> visited(gl.rsize(), false);

            while (true)
            {
                idx = node.top();

                if (gl.has_edge(idx,y))
                {
                    node.push(y);
                    break;
                }

                if (last.top() != gl.edges(idx).end())
                {
                    size_t n = *last.top();
                    ++last.top();

                    if (!visited[n])
                    {
                        visited[n] = true;
                        size_t k = m.left_of(n);
                        node.push(n);
                        node.push(k);
                        last.push(gl.edges(k).begin());
                    }          
                }
                else
                {
                    last.pop();
                    node.pop();
                    node.pop();
                }
            }

            while (!node.empty())
            {
                size_t j = node.top();
                node.pop();
                size_t i = node.top();
                node.pop();
                m.delete_edge_left(i);
                if (!m.add_edge(i,j))
                {
                    throw i3d::InternalException("Kuhn-Munkres algorithm: Internal error "
                        "in ExpandMatching method.");
                }
            }
        }

        /******************************************************************************/

        template <class DT>
        void KuhnMunkres(size_t n, const std::vector<DT> &weights, std::vector<size_t> &res)
        {
            if (weights.size() != n * n)
            {
                throw i3d::InternalException("Kuhn-Munkres algorithm: The size of the weights"
                    " matrix does not match parameter n.");
            }

            res.resize(n);

            if (!n)
            {
                return;
            }

            // vertex labels
            std::vector<DT> left(n); 
            std::vector<DT> right(n);

            // initial vertex labeling
            for (size_t i = 0; i < n; ++i)
            {
                // Row maximum
                left[i] = *std::max_element(&weights[i * n], &weights[(i + 1) * n]);
                right[i] = 0;
            }

            // Subgraph of input graph. It has the same number of vertices but
            // Only edges having weight equal to the sum of weights of their
            // vertices are included in the subgraph.
            BipartiteGraph gl;
            EqualitySubgraph(left, right, weights, gl);

            // vertex matching
            Matching m(n);
            ArbitraryMatching(gl, m);  

            std::set<size_t> S, T, J, JdiffT;
            DT alpha;
            size_t x, y, z;

            while (!m.is_complete())
            {
                x = m.get_unmatched();

                S.clear();
                T.clear();
                J.clear();

                S.insert(x);
                Enlarge_J(x, gl, J);
                JdiffT = J;

                do {
                    if (T == J)
                    {
                        std::set<size_t> Tc;
                        Complement(T, n, Tc);
                        
                        // Update labels
                        alpha = GetAlpha(left, right, S, Tc, n, weights);
                        for (std::set<size_t>::const_iterator si = S.begin(); si != S.end(); ++si)
                        {
                            left[*si] -= alpha;
                        }            
                        for (std::set<size_t>::const_iterator si = T.begin(); si != T.end(); ++si)
                        {
                            right[*si] += alpha;
                        }                        
                        
                        // Update subgraph
                        for (std::set<size_t>::const_iterator s = S.begin(); s != S.end(); ++s)
                        {
                            for (std::set<size_t>::const_iterator t = Tc.begin(); t != Tc.end(); ++t)
                            {
                                if (left[*s] + right[*t] - weights[(*s) * n + (*t)] <= DT(SMALL_DIFF))
                                {
                                    gl.add_edge(*s,*t);
                                    J.insert(*t);
                                    JdiffT.insert(*t);
                                }
                            }
                        }

                        // do gl pribudou pouze hrany, ktere vychazeji z mnoziny S
                        // Jejich konce jsou vlozeny do mnoziny J
                        // alespon jedna nova hrana vychazejici z mnoziny S vzdy existuje 
                    }
                    // zde je T vzdy podmnozinou J => J - T != 0 => JdiffT.size() > 0;
                    y = *JdiffT.begin();
                    z = m.left_of(y);

                    if (z == m.unused()) // y not used in matching
                        break;

                    // y is matched with z in matching m
                    S.insert(z);
                    T.insert(y);
                    JdiffT.erase(y);
                } while (true);

                ExpandMatching(m, gl, x, y);
            }

            res = m.matching();
        }

        /******************************************************************************/

        // Explicit instanitations
        template I3D_ALGO_EXPORT void KuhnMunkres(size_t n, const std::vector<float> &weights, 
            std::vector<size_t> &res);
        template I3D_ALGO_EXPORT void KuhnMunkres(size_t n, const std::vector<double> &weights, 
            std::vector<size_t> &res);

        /******************************************************************************/
    }
}
