/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** @file 
    @brief Point-based image registration methods.

    Based on the original code by Petr Matula.

    @author Ondrej Danek (xdanek2@fi.muni.cz)
    @date 2010
*/ 

#include <map>
#include "RegistrationKuhnMunkres.h"
#include "RegistrationPointBased.h"

#ifdef WITH_LAPACK
    /// LAPACK function for calculating eigenvalues and eigenvectors
    extern "C" void dsyev_(const char *jobz, const char *uplo, const int *n,
	    double *a, const int *lda, double *w, double *work, const int *lwork,
	    int *info);
#endif

namespace i3d
{
    namespace Registration
    {
        /******************************************************************************/

        static void MatchingPairsSupport(size_t i, size_t a, const PointSet &p, 
            const PointSet &q, double bin_width, double scale_eps, double zdiff_eps,
            size_t &mps, double &angle)
        {
            std::map<size_t, size_t> accum;
            std::vector<double> d_i;
            std::vector<double> angle_i;
            std::vector<double> u_z;
            std::vector<double> d_a;
            std::vector<double> angle_a;
            std::vector<double> v_z;

            size_t m = p.size();
            size_t n = q.size();

            d_i.resize(m);
            angle_i.resize(m);
            u_z.resize(m);

            for (size_t j=0; j < m; j++)
            {
                if(j != i)
                {
                    Vector3d<double> u = p[j] - p[i];
                    d_i[j] = Norm(u);
                    angle_i[j] = (atan2(u.y, u.x) * 180.0) / M_PI;
                    u_z[j] = u.z;
                }
            }

            d_a.resize(n);
            angle_a.resize(n);
            v_z.resize(n);

            for (size_t b=0; b < n; b++)
            {
                if (b != a)
                {
                    Vector3d<double> v = q[b] - q[a];
                    d_a[b] = Norm(v);
                    angle_a[b] = (atan2(v.y, v.x) * 180.0) / M_PI;
                    v_z[b] = v.z;
                }
            }

            for (size_t j = 0; j < m; j++)
            {
                if(j != i)
                {
                    for (size_t b=0; b < n; b++)
                    {
                        if (b != a)
                        {
                            double scale = d_i[j] / d_a[b];
                            double zdiff = v_z[b] - u_z[j];
                            if ((fabs(scale - 1.0) < scale_eps) && (fabs(zdiff) < zdiff_eps))
                            {
                                accum[(int)((angle_i[j] - angle_a[b]) / bin_width)]++;
                            }
                        }
                    }
                }
            }

            mps = 0;
            std::map<size_t, size_t>::const_iterator it;
            for (it=accum.begin(); it!=accum.end(); it++)
            {
                if (it->second > mps)
                {
                    mps = it->second;
                    angle = double(it->first) * bin_width;
                }
            }
        }

        /******************************************************************************/

        static size_t MaximalMatching(const PointSet &p, const PointSet &q, size_t &best_i, 
            size_t &best_a, double &best_angle, double bin_width, double scale_eps,
            double zdiff_eps)
        {
            best_i = 0;
            best_a = 0;
            best_angle = 0.0;

            size_t m = p.size();
            size_t n = q.size();
            size_t best_mps = 0;
            size_t k = m;

            for (size_t i = 0; i < m; i++)
            {
                for (size_t a = 0; a < n; a++)
                {
                    double angle;
                    size_t mps;

                    MatchingPairsSupport(i, a, p, q, bin_width, scale_eps, zdiff_eps, mps, angle);

                    if (best_mps < mps)
                    {
                        best_i = i;
                        best_a = a;
                        best_mps = mps;
                        best_angle = angle;
                    }

                    if (best_mps >= (k - 1))
                    {
                        return k;
                    }
                }

                k--; 
            }

            return k;
        }

        /******************************************************************************/

        void I3D_ALGO_EXPORT FindPointMatching(const PointSet &p, const PointSet &q, PointMatching &m,
            double max_dist, double bin_width, double scale_eps, double zdiff_eps)
        {
            // Coarse matching
            size_t best_i, best_a;
            double best_angle;
            
            MaximalMatching(p, q, best_i, best_a, best_angle, bin_width, scale_eps, zdiff_eps);

            // Transform point set q to the coordinate system of point set p
            std::vector<i3d::Vector3d<double> > q_in_p(q.size());

            double basin = sin((best_angle * M_PI) / 180.0);
            double bacos = cos((best_angle * M_PI) / 180.0);

            for (size_t i = 0; i < q.size(); i++)
            {
                // Translate point according to the centre of rotation
                q_in_p[i] = q[i] - q[best_a];

                // Rotate point by best_angle
                q_in_p[i].x = bacos * q_in_p[i].x + basin * q_in_p[i].y;
                q_in_p[i].y = bacos * q_in_p[i].y - basin * q_in_p[i].x;
                
                // Align point with set p
                q_in_p[i] += p[best_i];
            }

            // Initialization of weight matrix
            size_t w_size = p.size() + q.size();
            std::vector<double> w(w_size * w_size, 0.0);

            // Distances computed in Cartesians coordinations
            for (size_t i = 0; i < p.size(); i++)
            {
                for (size_t j = 0; j < q_in_p.size(); j++)
                {
                    w[i * w_size + j] = max_dist - Norm(q_in_p[j] - p[i]);
                }
            }

            // Fine matching using Kuhn Munkres
            std::vector<size_t> qr;            
            KuhnMunkres(w_size, w, qr);

            // Create matching
            m.clear();

            for (size_t i = 0; i < p.size(); ++i)
            {
                if (qr[i] < q.size())
                {
                    m.push_back(std::pair<size_t,size_t>(i, qr[i]));
                }
            }
        }

        /******************************************************************************/

        static void QuaternionToRotationMatrix(const double q[4], PointTransformation &mat)
        {
            double q00 = q[0]*q[0];
            double q11 = q[1]*q[1];
            double q22 = q[2]*q[2];
            double q33 = q[3]*q[3];

            double q01 = q[0]*q[1];
            double q02 = q[0]*q[2];
            double q03 = q[0]*q[3];
            double q12 = q[1]*q[2];
            double q13 = q[1]*q[3];
            double q23 = q[2]*q[3];

            mat[0][0] = q00 + q11 - q22 - q33;
            mat[0][1] = 2 * ( q12 + q03 );
            mat[0][2] = 2 * ( q13 - q02 );
            mat[0][3] = 0.0;

            mat[1][0] = 2 * ( q12 - q03 );
            mat[1][1] = q00 + q22 - q11 - q33;
            mat[1][2] = 2 * ( q23 + q01 );
            mat[1][3] = 0.0;

            mat[2][0] = 2 * ( q13 + q02 );
            mat[2][1] = 2 * ( q23 - q01 );
            mat[2][2] = q00 + q33 - q11 - q22;
            mat[2][3] = 0.0;

            mat[3][0] = 0.0;
            mat[3][1] = 0.0;
            mat[3][2] = 0.0;
            mat[3][3] = 1.0;
        }

        /******************************************************************************/

        static void AbsoluteOrientationProblem(const PointSet &ps1, const PointSet &ps2,
            PointTransformation &t)
        {
#ifdef WITH_LAPACK
            if (ps1.size() != ps2.size())
            {
                throw i3d::InternalException("AbsoluteOrientationProblem: Point sets must have "
                    "the same size.");
            }

            // Compute centroids
            i3d::Vector3d<double> pm1(0), pm2(0);

            for (size_t i = 0; i < ps1.size(); i++)
            {
                pm1 += ps1[i];
                pm2 += ps2[i];
            }

            pm1 /= double(ps1.size());
            pm2 /= double(ps2.size());

            // Compute cross-covariance matrix
            double cc[3][3];

            for (size_t i = 0; i < 3; ++i)
            {
                for (size_t j = 0; j < 3; ++j)
                {
                    cc[i][j] = 0;
                }
            }

            for (size_t k = 0; k < ps1.size(); ++k)
            {
                cc[0][0] += ps1[k].x * ps2[k].x;
                cc[0][1] += ps1[k].x * ps2[k].y;
                cc[0][2] += ps1[k].x * ps2[k].z;
                cc[1][0] += ps1[k].y * ps2[k].x;
                cc[1][1] += ps1[k].y * ps2[k].y;
                cc[1][2] += ps1[k].y * ps2[k].z;
                cc[2][0] += ps1[k].z * ps2[k].x;
                cc[2][1] += ps1[k].z * ps2[k].y;
                cc[2][2] += ps1[k].z * ps2[k].z;
            }

            for (size_t i = 0; i < 3; ++i)
            {
                for (size_t j = 0; j < 3; ++j)
                {
                    cc[i][j] = cc[i][j] / double(ps1.size()) - pm1[i] * pm2[j];
                }
            }

            // Symetric 4x4 matrix computed from cc
            double Q[4][4];
            Q[0][0] = cc[0][0] + cc[1][1] + cc[2][2];
            Q[1][0] = cc[1][2] - cc[2][1];
            Q[2][0] = cc[2][0] - cc[0][2];
            Q[3][0] = cc[0][1] - cc[1][0];

            Q[1][1] = cc[0][0] - cc[1][1] - cc[2][2];
            Q[2][1] = cc[0][1] + cc[1][0];
            Q[3][1] = cc[0][2] + cc[2][0];

            Q[2][2] = -cc[0][0] + cc[1][1] - cc[2][2];
            Q[3][2] = cc[1][2] + cc[2][1];

            Q[3][3] = -cc[0][0] - cc[1][1] + cc[2][2];

            // Compute eigen values and eigen vectors of Q
            double ev[4], work[16];
            int info, dim = 4, lda = 4, lwork = 16;
            dsyev_("V", "U", &dim, (double *)Q, &lda, ev, work, &lwork, &info);

            // Rotation quaternion is the last eigenvector (largest eigenvalue)
            QuaternionToRotationMatrix(&Q[3][0], t);

            // Compute the scale factor - using symmetrical expression
            // for the error term
            double s1 = 0.0, s2 = 0.0;

            for (size_t i = 0; i < ps1.size(); i++)
            {
                s1 += Norm2(ps1[i] - pm1);
                s2 += Norm2(ps2[i] - pm2);
            }

            double s = sqrt(s1 / s2);

            // Compose scale and rotation
            for (size_t i = 0; i < 3; i++)
            {
                for (size_t j = 0; j < 3; j++)
                {
                    t[i][j] *= s;
                }
            }

            // Rotate and scale the center of mass of the second point set 
            double m3[3];
            for (size_t i = 0; i < 3; i++)
            {
                m3[i] = t[i][0] * pm2.x + t[i][1] * pm2.y + t[i][2] * pm2.z;
            }

            // Compute the translation vector
            t[0][3] = pm1.x - m3[0];
            t[1][3] = pm1.y - m3[1];
            t[2][3] = pm1.z - m3[2];
#else // WITH_LAPACK
            throw i3d::InternalException("AbsoluteOrientationProblem: Lapack library required.");
#endif
        }

        /******************************************************************************/

        void I3D_ALGO_EXPORT FindPointTransformation(const PointSet &p, const PointSet &q, 
            const PointMatching &m, PointTransformation &t)
        {
		    PointSet np1(m.size()), np2(m.size());
		    
            // Create corresponding Point Sets
            for (size_t k = 0; k < m.size(); ++k)
            {
                np1[k] = p[m[k].first];
                np2[k] = q[m[k].second];
            } 

            // Solve least-square alignment problem
		    AbsoluteOrientationProblem(np1, np2, t);
        }

        /******************************************************************************/

    }
}
