/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file lbptop.cc
   \brief 3D LBP-TOP features.
	
	This file contains functions for calculating 3D LBP-TOP features.

	[1] G. Zhao and M. Pietikinen. Dynamic Texture Recognition Using Local
	Binary Patterns with an Application to Facial Expressions. IEEE Transactions
	on Pattern Analysis and Machine Intelligence, 29(6):915928, June
	2007.
	
	M. Pietikinen, A. Hadid, G. Zhao, and T. Ahonen. Computer Vision Using
	Local Binary Patterns. Computational imaging and vision. Springer
	Verlag, London, 2011.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include <fstream>
#include <math.h>

#include "descriptors/lbptop.h"
#include "transform.h"

// default values from [1]
#define P 8			// number of sampling points around the circle
#define R 1			// radius


using namespace std;

namespace i3d {


template <class T> void ComputeLBP_TOP(const Image3d<T> &img_orig, vector<double> &feature_vector)
{
	Image3d<T> img = img_orig;

	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();
	if ((res.x != res.y) || (res.x != res.z))
	{
		float max_res = std::max(std::max(res.x, res.y), res.z);
		Vector3d<float> new_res = max_res;
		ResampleToDesiredResolution(img, new_res, LANCZOS);
	}

	int hist_len = (int) pow(2.0, P);   // length of the histogram
	int *resultXY = new int[hist_len];  // 
	int *resultXZ = new int[hist_len];  // three values for three orthogonal planes 
	int *resultYZ = new int[hist_len];  // 

	// clear the variables
	for (int i = 0; i < hist_len; i++)	
	{
		resultXY[i] = 0; resultXZ[i] = 0; resultYZ[i] = 0;
	}

	for (unsigned z = R; z < img.GetSizeZ() - R; z++)
	{
		for (unsigned y = R; y < img.GetSizeY() - R; y++)
		{
			for (unsigned x = R; x < img.GetSizeX() - R; x++)
			{
				if (img.GetVoxel(x,y,z) != 0)
				{
					// three values, one for each LBP pattern in particular 2D plane
					int patternXY = 0, patternXZ = 0, patternYZ = 0;
					for (int p = 0; p < P; p++)
					{
						// finding x and y coordinates of point in the neighbourhood
						double xp = x + (((R*cos(2*M_PI*p/P) > -0.05) && (R*cos(2*M_PI*p/P) < 0.05)) ? 0 : R*cos(2*M_PI*p/P));
						double yp = y - (((R*sin(2*M_PI*p/P) > -0.05) && (R*sin(2*M_PI*p/P) < 0.05)) ? 0 : R*sin(2*M_PI*p/P));
		
						// if we are right in the centre of the pixel, we can take its value
						if ((xp == (int)xp) || (yp == (int)yp))
						{
							patternXY += ((img.GetVoxel((int)(xp + 0.1), (int)(yp + 0.1), z) < img.GetVoxel(x, y, z)) ? 0 : 1) << p;
						}
						else  // if we are not, we need to use bilinear interpolation to get the value
						{
							int x1 = (int)floor(xp), x2 = (int)ceil(xp), y1 = (int)floor(yp), y2 = (int)ceil(yp);
							patternXY += (((img.GetVoxel(x1,y1,z)*(x2-xp)*(y2-yp) +
										img.GetVoxel(x2,y1,z)*(xp-x1)*(y2-yp) +
										img.GetVoxel(x1,y2,z)*(x2-xp)*(yp-y1) +
										img.GetVoxel(x2,y2,z)*(xp-x1)*(yp-y1)) < img.GetVoxel(x, y, z)) ? 0 : 1) << p;
						}

						// finding z coordinate of point in the neighbourhood
						double zp = z - (((R*sin(2*M_PI*p/P) > -0.05) && (R*sin(2*M_PI*p/P) < 0.05)) ? 0 : R*sin(2*M_PI*p/P));

						// if we are right in the centre of the pixel, we can take its value
						if ((xp == (int)xp) || (zp == (int)zp))
						{
							patternXZ += ((img.GetVoxel((int)(xp + 0.1), y, (int)(zp + 0.1)) < img.GetVoxel(x, y, z)) ? 0 : 1) << p;
						}
						else  // if we are not, we need to use bilinear interpolation to get the value
						{
							int x1 = (int)floor(xp), x2 = (int)ceil(xp), z1 = (int)floor(zp), z2 = (int)ceil(zp);
							patternXZ += (((img.GetVoxel(x1,y,z1)*(x2-xp)*(z2-zp) +
										img.GetVoxel(x2,y,z1)*(xp-x1)*(z2-zp) +
										img.GetVoxel(x1,y,z2)*(x2-xp)*(zp-z1) +
										img.GetVoxel(x2,y,z2)*(xp-x1)*(zp-z1)) < img.GetVoxel(x, y, z)) ? 0 : 1) << p;
						}

						// finding y coordinate of point in the neighbourhood
						yp = y + (((R*cos(2*M_PI*p/P) > -0.05) && (R*cos(2*M_PI*p/P) < 0.05)) ? 0 : R*cos(2*M_PI*p/P));
		
						// if we are right in the centre of the pixel, we can take its value
						if ((yp == (int)yp) || (zp == (int)zp))
						{
							patternYZ += ((img.GetVoxel(x, (int)(yp + 0.1), (int)(zp + 0.1)) < img.GetVoxel(x, y, z)) ? 0 : 1) << p;
						}
						else  // if we are not, we need to use bilinear interpolation to get the value
						{
							int y1 = (int)floor(yp), y2 = (int)ceil(yp), z1 = (int)floor(zp), z2 = (int)ceil(zp);
							patternYZ += (((img.GetVoxel(x,y1,z1)*(y2-yp)*(z2-zp) +
										img.GetVoxel(x,y2,z1)*(yp-y1)*(z2-zp) +
										img.GetVoxel(x,y1,z2)*(y2-yp)*(zp-z1) +
										img.GetVoxel(x,y2,z2)*(yp-y1)*(zp-z1)) < img.GetVoxel(x, y, z)) ? 0 : 1) << p;
						}
					}
					resultXY[patternXY]++; resultXZ[patternXZ]++; resultYZ[patternYZ]++;
				}
			}
		}	
	}

	// for normalization
	int sum_1 = 0, sum_2 = 0, sum_3 = 0;
	for (int i = 0; i < hist_len; i++)
	{
		sum_1 += resultXY[i]; sum_2 += resultXZ[i]; sum_3 += resultYZ[i];
	}

	// form the normalized feature vector
	for (int i = 0; i<hist_len; i++) feature_vector.push_back((double)(resultXY[i])/sum_1);
	for (int i = 0; i<hist_len; i++) feature_vector.push_back((double)(resultXZ[i])/sum_2);
	for (int i = 0; i<hist_len; i++) feature_vector.push_back((double)(resultYZ[i])/sum_3);
	
	delete []resultXY;
	delete []resultXZ;
	delete []resultYZ;  
}

//-----------------------------------------------------------------------
// explicit instantiation
//-----------------------------------------------------------------------

template I3D_ALGO_EXPORT void ComputeLBP_TOP<GRAY8>(const Image3d<GRAY8> &, vector<double> &);
template I3D_ALGO_EXPORT void ComputeLBP_TOP<GRAY16>(const Image3d<GRAY16> &, vector<double> &);

}
