/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * lab_nha_result_creator.cpp
 *
 * Result creator used in thelabelled version of the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "lab_nha_result_creator.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                 Class Lab_NHA_ResultCreator                 //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Lab_NHA_ResultCreator::Labels(Type::LabVoxel *labels) const
		{
			// index of the starting grid point
			Type::Size index = m_grid.m_sx + 1;

			// the first grid point in given labeled image
			Type::LabVoxel *ptr = labels;

			// implicit function at the starting grid point
			const int *iptr = m_grid.m_implicit_function.ptr(index);

			// label at the starting grid point
			const Type::LabVoxel *lptr = m_lab.ptr(index);

			for (Type::Size y = 2; y < m_grid.m_sy; ++y)
			{
				for (Type::Size x = 2; x < m_grid.m_sx; ++x)
				{
					// copy a label to the labelled image
					*ptr = *lptr;

					// move to the next grid point
					++lptr;
					++iptr;
					++ptr;
				}

				// move to the next row
				lptr += 2;
				iptr += 2;
			}
		}

		/**********************************************************************/

		void Lab_NHA_ResultCreator::LabelsInDomain(Type::LabVoxel *labels) const
		{
			// index of the starting grid point
			Type::Size index = m_grid.m_sx + 1;

			// the first grid point in given labeled image
			Type::LabVoxel *ptr = labels;

			// implicit function at the starting grid point
			const int *iptr = m_grid.m_implicit_function.ptr(index);

			// label at the starting grid point
			const Type::LabVoxel *lptr = m_lab.ptr(index);

			for (Type::Size y = 2; y < m_grid.m_sy; ++y)
			{
				for (Type::Size x = 2; x < m_grid.m_sx; ++x)
				{
					// check whether a grid point is in the domain
					if (*iptr != 5)
					{
						// copy a label to the labelled image
						*ptr = *lptr;
					}

					// move to the next grid point
					++lptr;
					++iptr;
					++ptr;
				}

				// move to the next row
				lptr += 2;
				iptr += 2;
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                Class Lab_NHA_ResultCreator3D                //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Lab_NHA_ResultCreator3D::Labels(Type::LabVoxel *labels) const
		{
			// index of the starting grid point
			Type::Size index = m_grid.m_slice + m_grid.m_sx + 1;

			// the first grid point in given labeled image
			Type::LabVoxel *ptr = labels;

			// implicit function at the starting grid point
			const int *iptr = m_grid.m_implicit_function.ptr(index);

			// label at the starting grid point
			const Type::LabVoxel *lptr = m_lab.ptr(index);

			// shift to the next slice
			Type::Size shift = 2 * m_grid.m_sx;

			for (Type::Size z = 2; z < m_grid.m_sz; ++z)
			{
				for (Type::Size y = 2; y < m_grid.m_sy; ++y)
				{
					for (Type::Size x = 2; x < m_grid.m_sx; ++x)
					{
						// copy a label to the labelled image
						*ptr = *lptr;

						// move to the next grid point
						++lptr;
						++iptr;
						++ptr;
					}

					// move to the next row
					lptr += 2;
					iptr += 2;

				}

				// move to the next slice
				lptr += shift;
				iptr += shift;
			}
		}

		/**********************************************************************/

		void Lab_NHA_ResultCreator3D::LabelsInDomain(Type::LabVoxel *labels) const
		{
			// index of the starting grid point
			Type::Size index = m_grid.m_slice + m_grid.m_sx + 1;

			// the first grid point in given labeled image
			Type::LabVoxel *ptr = labels;

			// implicit function at the starting grid point
			const int *iptr = m_grid.m_implicit_function.ptr(index);

			// label at the starting grid point
			const Type::LabVoxel *lptr = m_lab.ptr(index);

			// shift to the next slice
			Type::Size shift = 2 * m_grid.m_sx;

			for (Type::Size z = 2; z < m_grid.m_sz; ++z)
			{
				for (Type::Size y = 2; y < m_grid.m_sy; ++y)
				{
					for (Type::Size x = 2; x < m_grid.m_sx; ++x)
					{
						// check whether a grid point is in the domain
						if (*iptr != 5)
						{
							// copy a label to the labelled image
							*ptr = *lptr;
						}

						// move to the next grid point
						++lptr;
						++iptr;
						++ptr;
					}

					// move to the next row
					lptr += 2;
					iptr += 2;
				}

				// move to the next slice
				lptr += shift;
				iptr += shift;
			}
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

