/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nha_advection_force.cc
 *
 * Advection force used in the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "nha_advection_force.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                  Class NHA_AdvectionForce                   //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Type::Prec NHA_AdvectionForce::Get(Type::Size i) const
		{
			Type::Size backward_index = i - 1;
			Type::Size forward_index  = i + 1;

			int tmp = m_grid.m_implicit_function[forward_index] - m_grid.m_implicit_function[backward_index];
			Type::Prec imp_grad = (tmp < -2 || tmp > 2) ? Type::Prec(0) : tmp;

			Type::Prec imp_grad_mag = Sqr(imp_grad) * m_denom_x;
			Type::Prec advection = imp_grad * (m_grad_mag[forward_index] - m_grad_mag[backward_index]) * m_denom_x;

			backward_index = i - m_grid.m_sx;
			forward_index  = i + m_grid.m_sx;

			tmp = m_grid.m_implicit_function[forward_index] - m_grid.m_implicit_function[backward_index];
			imp_grad = (tmp < -2 || tmp > 2) ? Type::Prec(0) : tmp;

			imp_grad_mag += (Sqr(imp_grad) * m_denom_y);
			advection += (imp_grad * (m_grad_mag[forward_index] - m_grad_mag[backward_index]) * m_denom_y);

			return PosDiv(advection, Sqrt(imp_grad_mag));
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                 Class NHA_AdvectionForce3D                  //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Type::Prec NHA_AdvectionForce3D::Get(Type::Size i) const
		{
			Type::Size backward_index = i - 1;
			Type::Size forward_index  = i + 1;

			int tmp = m_grid.m_implicit_function[forward_index] - m_grid.m_implicit_function[backward_index];
			Type::Prec imp_grad = (tmp < -2 || tmp > 2) ? Type::Prec(0) : tmp;

			Type::Prec imp_grad_mag = Sqr(imp_grad) * m_denom_x;
			Type::Prec advection = imp_grad * (m_grad_mag[forward_index] - m_grad_mag[backward_index]) * m_denom_x;

			backward_index = i - m_grid.m_sx;
			forward_index  = i + m_grid.m_sx;

			tmp = m_grid.m_implicit_function[forward_index] - m_grid.m_implicit_function[backward_index];
			imp_grad = (tmp < -2 || tmp > 2) ? Type::Prec(0) : tmp;

			imp_grad_mag += (Sqr(imp_grad) * m_denom_y);
			advection += (imp_grad * (m_grad_mag[forward_index] - m_grad_mag[backward_index]) * m_denom_y);

			backward_index = i - m_grid.m_slice;
			forward_index  = i + m_grid.m_slice;

			tmp = m_grid.m_implicit_function[forward_index] - m_grid.m_implicit_function[backward_index];
			imp_grad = (tmp < -2 || tmp > 2) ? Type::Prec(0) : tmp;

			imp_grad_mag += (Sqr(imp_grad) * m_denom_z);
			advection += (imp_grad * (m_grad_mag[forward_index] - m_grad_mag[backward_index]) * m_denom_z);

			return PosDiv(advection, Sqrt(imp_grad_mag));
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

