/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nha_curvature.cc
 *
 * Curvature used in the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "nha_curvature.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                     Class NHA_Curvature                     //
		//                                                             //
		/////////////////////////////////////////////////////////////////
				
		/**********************************************************************/

		void NHA_Curvature::Update(Type::Size i, int value)
		{
			// compute limits of the circular neighbourhood in the x axis
			Type::Size min_x, max_x;
			m_circular_neigh -> LimitsX(i, min_x, max_x);

			// compute limits of the circular neighbourhood in the y axis
			Type::Size min_y, max_y;
			m_circular_neigh -> LimitsY(i, min_y, max_y);

			// shift to the next row of the circular neighbourhood
			Type::Size shift = m_grid.m_sx - max_x + min_x - 1;

			// index of the first grid point in the circular neighbourhood
			Type::Size index = min_y * m_grid.m_sx + min_x;

			// curvature at the first grid point in the circular neighbourhood
			int *ptr = m_grid.m_curvature.ptr(index);

			// position of a grid point on the heap
			Type::Size pos;

			// go through the circular neighbourhood and update the curvature at grid points on the heap
			for (; min_y <= max_y; ++min_y)
			{
				for (Type::Size x = min_x; x <= max_x; ++x)
				{
					// get position on the heap
					pos = m_heap.GetPosition(index);

					// check whether a grid point is on the heap
					if (pos != MinHeap<Type::Prec, NHA_Node>::NOT_ON_HEAP)
					{
						// update the curvature by given value
						*ptr += value;

						// update the heap
						m_heap.Update(pos, m_speed_function -> Get(index));
					}

					++index;
					++ptr;
				}

				// move to the next row of the circular neighbourhood
				index += shift;
				ptr += shift;
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                    Class NHA_Curvature3D                    //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void NHA_Curvature3D::Update(Type::Size i, int value)
		{
			// compute limits of the circular neighbourhood in the x axis
			Type::Size min_x, max_x;
			m_circular_neigh -> LimitsX(i, min_x, max_x);

			// compute limits of the circular neighbourhood in the y axis
			Type::Size min_y, max_y;
			m_circular_neigh -> LimitsY(i, min_y, max_y);

			// compute limits of the circular neighbourhood in the z axis
			Type::Size min_z, max_z;
			m_circular_neigh -> LimitsZ(i, min_z, max_z);

			// shift to the next row of the circular neighbourhood
			Type::Size row_shift = m_grid.m_sx - max_x + min_x - 1;

			// shift to the next slice of the circular neighbourhood
			Type::Size slice_shift = m_grid.m_slice - (max_y - min_y + 1) * m_grid.m_sx;

			// index of the first grid point in the circular neighbourhood
			Type::Size index = min_z * m_grid.m_slice + min_y * m_grid.m_sx + min_x;

			// curvature at the first grid point in the circular neighbourhood
			int *ptr = m_grid.m_curvature.ptr(index);

			// position of a grid point on the heap
			Type::Size pos;

			// go through the circular neighbourhood and update the curvature at grid points on the heap
			for (; min_z <= max_z; ++min_z)
			{
				for (Type::Size y = min_y; y <= max_y; ++y)
				{
					for (Type::Size x = min_x; x <= max_x; ++x)
					{
						// get position on the heap
						pos = m_heap.GetPosition(index);

						// check whether a grid point is on the heap
						if (pos != MinHeap<Type::Prec, NHA_Node>::NOT_ON_HEAP)
						{
							// update the curvature by given value
							*ptr += value;

							// update the heap
							m_heap.Update(pos, m_speed_function -> Get(index));
						}

						++index;
						++ptr;
					}

					// move to the next row of the circular neighbourhood
					index += row_shift;
					ptr += row_shift;
				}

				// move to the next slice of the circular neighbourhood
				index += slice_shift;
				ptr += slice_shift;
			}
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

