/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * tp_glob_reg_nha_routine.cc
 *
 * Topology-preserving routine for the global region-based version of the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "tp_glob_reg_nha_routine.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                Class TP_Glob_Reg_NHA_Routine                //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void TP_Glob_Reg_NHA_Routine::PerformIteration()
		{
			// compute the speed at each interface point
			Type::Size num = InterfaceSize();
			Buffer<Type::Prec> speed(num);
			std::vector<Type::Size> negative, positive;
			for (Type::Size i = 0; i < num; ++i)
			{
				m_curvature -> Estimate(m_base -> m_interface[i]);
				speed[i] = m_speed_function -> Get(m_base -> m_interface[i]);

				if (speed[i] > Type::Prec(0))
				{
					positive.push_back(i);
				}
				else
				{
					negative.push_back(i);
				}
			}

			std::vector<Type::Size> new_interface;
			Type::Size num_neigh = 0;
			const Type::Size *neigh = NULL;

			// move the interface points locally
			for (Type::Size i = 0; i < negative.size(); ++i)
			{
				if (m_simple_point_inspector -> Verify(m_base -> m_interface[negative[i]]))
				{
					neigh = m_neighbourhood -> InternalNeighbours(m_base -> m_interface[negative[i]], num_neigh);

					m_base -> m_implicit_function[m_base -> m_interface[negative[i]]] = 1;
					m_obj_manager -> SwapForeground(m_base -> m_interface[negative[i]]);

					for (Type::Size n = 0; n < num_neigh; ++n)
					{
						m_base -> m_implicit_function[neigh[n]] = 0;
						new_interface.push_back(neigh[n]);
					}
				}
				else
				{ 
					new_interface.push_back(m_base -> m_interface[negative[i]]);
				}
			}

			bool is_ok;

			for (Type::Size i = 0; i < positive.size(); ++i)
			{
				is_ok = true;
				neigh = m_neighbourhood -> ExternalNeighbours(m_base -> m_interface[positive[i]], num_neigh);

				for (Type::Size n = 0; n < num_neigh; ++n)
				{
					if (m_simple_point_inspector -> Verify(neigh[n]))
					{
						m_base -> m_implicit_function[neigh[n]] = 0;
						m_obj_manager -> SwapBackground(neigh[n]);
						new_interface.push_back(neigh[n]);
					}
					else
					{
						is_ok = false;
					}
				}

				if (is_ok)
				{
					m_base -> m_implicit_function[m_base -> m_interface[positive[i]]] = -1;
				}
				else
				{
					new_interface.push_back(m_base -> m_interface[positive[i]]);
				}
			}

			m_base -> m_interface = new_interface;

			// synchronize foreground and background objects
			m_obj_manager -> Synchronize();
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

