/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2018   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

//------------------------------------------------------------------------

/*
 * FILE: skeleton.cc
 *
 * An implementation of the algorithm from Lee et al. "Building skeleton models via 3-D 
 * medial surface/axis thinning algorithms. Computer Vision, Graphics, and 
 * Image Processing, 56(6):462–478, 1994." Adopted from the ITK version from
 * Hanno Homann:  http://hdl.handle.net/1926/1292 and FIJI plugin from 
 * Ignacio Arganda-Carreras: https://imagej.net/Skeletonize3D
 *
 * Imported and modified by: David Svoboda <svoboda@fi.muni.cz> August 2018
 */

//------------------------------------------------------------------------

#include <image3d.h>
#include <neighbours.h>

#include "skeleton.h"

namespace i3d {
	
const int neighbours[] = 
{
	 -1,-1,-1,	 0,-1,-1,	 1,-1,-1,
	 -1, 0,-1,	 0, 0,-1,	 1, 0,-1,
	 -1, 1,-1,	 0, 1,-1,	 1, 1,-1,
	 -1,-1, 0,	 0,-1, 0,	 1,-1, 0,
	 -1, 0, 0,	 0, 0, 0,	 1, 0, 0,
	 -1, 1, 0,	 0, 1, 0,	 1, 1, 0,
	 -1,-1, 1,	 0,-1, 1,	 1,-1, 1,
	 -1, 0, 1,	 0, 0, 1,	 1, 0, 1,
	 -1, 1, 1,	 0, 1, 1,	 1, 1, 1
};

const int count = 27;

const i3d::Neighbourhood nb3D_full(neighbours, count);

void fillEulerLUT(int *LUT) 
{
		LUT[1]  =  1;
		LUT[3]  = -1;
		LUT[5]  = -1;
		LUT[7]  =  1;
		LUT[9]  = -3;
		LUT[11] = -1;
		LUT[13] = -1;
		LUT[15] =  1;
		LUT[17] = -1;
		LUT[19] =  1;
		LUT[21] =  1;
		LUT[23] = -1;
		LUT[25] =  3;
		LUT[27] =  1;
		LUT[29] =  1;
		LUT[31] = -1;
		LUT[33] = -3;
		LUT[35] = -1;
		LUT[37] =  3;
		LUT[39] =  1;
		LUT[41] =  1;
		LUT[43] = -1;
		LUT[45] =  3;
		LUT[47] =  1;
		LUT[49] = -1;
		LUT[51] =  1;

		LUT[53] =  1;
		LUT[55] = -1;
		LUT[57] =  3;
		LUT[59] =  1;
		LUT[61] =  1;
		LUT[63] = -1;
		LUT[65] = -3;
		LUT[67] =  3;
		LUT[69] = -1;
		LUT[71] =  1;
		LUT[73] =  1;
		LUT[75] =  3;
		LUT[77] = -1;
		LUT[79] =  1;
		LUT[81] = -1;
		LUT[83] =  1;
		LUT[85] =  1;
		LUT[87] = -1;
		LUT[89] =  3;
		LUT[91] =  1;
		LUT[93] =  1;
		LUT[95] = -1;
		LUT[97] =  1;
		LUT[99] =  3;
		LUT[101] =  3;
		LUT[103] =  1;

		LUT[105] =  5;
		LUT[107] =  3;
		LUT[109] =  3;
		LUT[111] =  1;
		LUT[113] = -1;
		LUT[115] =  1;
		LUT[117] =  1;
		LUT[119] = -1;
		LUT[121] =  3;
		LUT[123] =  1;
		LUT[125] =  1;
		LUT[127] = -1;
		LUT[129] = -7;
		LUT[131] = -1;
		LUT[133] = -1;
		LUT[135] =  1;
		LUT[137] = -3;
		LUT[139] = -1;
		LUT[141] = -1;
		LUT[143] =  1;
		LUT[145] = -1;
		LUT[147] =  1;
		LUT[149] =  1;
		LUT[151] = -1;
		LUT[153] =  3;
		LUT[155] =  1;

		LUT[157] =  1;
		LUT[159] = -1;
		LUT[161] = -3;
		LUT[163] = -1;
		LUT[165] =  3;
		LUT[167] =  1;
		LUT[169] =  1;
		LUT[171] = -1;
		LUT[173] =  3;
		LUT[175] =  1;
		LUT[177] = -1;
		LUT[179] =  1;
		LUT[181] =  1;
		LUT[183] = -1;
		LUT[185] =  3;
		LUT[187] =  1;
		LUT[189] =  1;
		LUT[191] = -1;
		LUT[193] = -3;
		LUT[195] =  3;
		LUT[197] = -1;
		LUT[199] =  1;
		LUT[201] =  1;
		LUT[203] =  3;
		LUT[205] = -1;
		LUT[207] =  1;

		LUT[209] = -1;
		LUT[211] =  1;
		LUT[213] =  1;
		LUT[215] = -1;
		LUT[217] =  3;
		LUT[219] =  1;
		LUT[221] =  1;
		LUT[223] = -1;
		LUT[225] =  1;
		LUT[227] =  3;
		LUT[229] =  3;
		LUT[231] =  1;
		LUT[233] =  5;
		LUT[235] =  3;
		LUT[237] =  3;
		LUT[239] =  1;
		LUT[241] = -1;
		LUT[243] =  1;
		LUT[245] =  1;
		LUT[247] = -1;
		LUT[249] =  3;
		LUT[251] =  1;
		LUT[253] =  1;
		LUT[255] = -1;
}
	
inline unsigned char indexOctantNEB(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;

		if( *(neighbors[2])==true )
			n |= 128;
		if( *(neighbors[1])==true )
			n |=  64;
		if( *(neighbors[11])==true )
			n |=  32;
		if( *(neighbors[10])==true )
			n |=  16;
		if( *(neighbors[5])==true )
			n |=   8;
		if( *(neighbors[4])==true )
			n |=   4;
		if( *(neighbors[14])==true )
			n |=   2;

		return n;
}

inline unsigned char indexOctantNWB(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;
		if( *(neighbors[0])==true )
			n |= 128;
		if( *(neighbors[9])==true )
			n |=  64;
		if( *(neighbors[3])==true )
			n |=  32;
		if( *(neighbors[12])==true )
			n |=  16;
		if( *(neighbors[1])==true )
			n |=   8;
		if( *(neighbors[10])==true )
			n |=   4;
		if( *(neighbors[4])==true )
			n |=   2;
		return n;
}

inline unsigned char indextOctantSEB(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;
		if( *(neighbors[8])==true )
			n |= 128;
		if( *(neighbors[7])==true )
			n |=  64;
		if( *(neighbors[17])==true )
			n |=  32;
		if( *(neighbors[16])==true )
			n |=  16;
		if( *(neighbors[5])==true )
			n |=   8;
		if( *(neighbors[4])==true )
			n |=   4;
		if( *(neighbors[14])==true )
			n |=   2;
		return n;
}

inline unsigned char indexOctantSWB(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;
		if( *(neighbors[6])==true )
			n |= 128;
		if( *(neighbors[15])==true )
			n |=  64;
		if( *(neighbors[7])==true )
			n |=  32;
		if( *(neighbors[16])==true )
			n |=  16;
		if( *(neighbors[3])==true )
			n |=   8;
		if( *(neighbors[12])==true )
			n |=   4;
		if( *(neighbors[4])==true )
			n |=   2;
		return n;
}

inline unsigned char indexOctantNEU(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;
		if( *(neighbors[20])==true )
			n |= 128;
		if( *(neighbors[23])==true )
			n |=  64;
		if( *(neighbors[19])==true )
			n |=  32;
		if( *(neighbors[22])==true )
			n |=  16;
		if( *(neighbors[11])==true )
			n |=   8;
		if( *(neighbors[14])==true )
			n |=   4;
		if( *(neighbors[10])==true )
			n |=   2;
		return n;
}

inline unsigned char indexOctantNWU(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;
		if( *(neighbors[18])==true )
			n |= 128;
		if( *(neighbors[21])==true )
			n |=  64;
		if( *(neighbors[9])==true )
			n |=  32;
		if( *(neighbors[12])==true )
			n |=  16;
		if( *(neighbors[19])==true )
			n |=   8;
		if( *(neighbors[22])==true )
			n |=   4;
		if( *(neighbors[10])==true )
			n |=   2;
		return n;
}

inline unsigned char indexOctantSEU(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;
		if( *(neighbors[26])==true )
			n |= 128;
		if( *(neighbors[23])==true )
			n |=  64;
		if( *(neighbors[17])==true )
			n |=  32;
		if( *(neighbors[14])==true )
			n |=  16;
		if( *(neighbors[25])==true )
			n |=   8;
		if( *(neighbors[22])==true )
			n |=   4;
		if( *(neighbors[16])==true )
			n |=   2;
		return n;
}

inline unsigned char indexOctantSWU(const std::vector<const bool*> &neighbors) 
{
		unsigned char n;
		n = 1;
		if( *(neighbors[24])==true )
			n |= 128;
		if( *(neighbors[25])==true )
			n |=  64;
		if( *(neighbors[15])==true )
			n |=  32;
		if( *(neighbors[16])==true )
			n |=  16;
		if( *(neighbors[21])==true )
			n |=   8;
		if( *(neighbors[22])==true )
			n |=   4;
		if( *(neighbors[12])==true )
			n |=   2;
		return n;
}
	
/**
 * Check if a point is Euler invariant
 * 
 * @param neighbors neighbor pixels of the point
 * @param LUT Euler LUT
 * @return true or false if the point is Euler invariant or not
 */
inline bool isEulerInvariant(const std::vector<const bool*> &neighbors, const int *LUT)
{
		// Calculate Euler characteristic for each octant and sum up
		int eulerChar = 0;
		unsigned char n;

		// Octant SWU
		n = indexOctantSWU(neighbors);
		eulerChar += LUT[n];
		
		// Octant SEU
		n = indexOctantSEU(neighbors);
		eulerChar += LUT[n];
		
		// Octant NWU
		n = indexOctantNWU(neighbors);
		eulerChar += LUT[n];
		
		// Octant NEU
		n = indexOctantNEU(neighbors);
		eulerChar += LUT[n];
		
		// Octant SWB
		n = indexOctantSWB(neighbors);
		eulerChar += LUT[n];
		
		// Octant SEB
		n = indextOctantSEB(neighbors);
		eulerChar += LUT[n];
		
		// Octant NWB
		n = indexOctantNWB(neighbors);
		eulerChar += LUT[n];
		
		// Octant NEB
		n = indexOctantNEB(neighbors);
		eulerChar += LUT[n];

		return (eulerChar == 0);
}

/**
 * This is a recursive method that calculates the number of connected
 * components in the 3D neighborhood after the center pixel would
 * have been removed.
 * 
 * @param octant
 * @param label
 * @param cube
 */
void octreeLabeling(int octant, int label, int *cube) 
{
		// check if there are points in the octant with value 1
		  if( octant==1 )
		  {
		  	// set points in this octant to current label
		  	// and recursive labeling of adjacent octants
		    if( cube[0] == 1 )
		      cube[0] = label;
		    if( cube[1] == 1 )
		    {
		      cube[1] = label;        
		      octreeLabeling( 2, label, cube);
		    }
		    if( cube[3] == 1 )
		    {
		      cube[3] = label;        
		      octreeLabeling( 3, label, cube);
		    }
		    if( cube[4] == 1 )
		    {
		      cube[4] = label;        
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 4, label, cube);
		    }
		    if( cube[9] == 1 )
		    {
		      cube[9] = label;        
		      octreeLabeling( 5, label, cube);
		    }
		    if( cube[10] == 1 )
		    {
		      cube[10] = label;        
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 5, label, cube);
		      octreeLabeling( 6, label, cube);
		    }
		    if( cube[12] == 1 )
		    {
		      cube[12] = label;        
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 5, label, cube);
		      octreeLabeling( 7, label, cube);
		    }
		  }
		  if( octant==2 )
		  {
		    if( cube[1] == 1 )
		    {
		      cube[1] = label;
		      octreeLabeling( 1, label, cube);
		    }
		    if( cube[4] == 1 )
		    {
		      cube[4] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 4, label, cube);
		    }
		    if( cube[10] == 1 )
		    {
		      cube[10] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 5, label, cube);
		      octreeLabeling( 6, label, cube);
		    }
		    if( cube[2] == 1 )
		      cube[2] = label;        
		    if( cube[5] == 1 )
		    {
		      cube[5] = label;        
		      octreeLabeling( 4, label, cube);
		    }
		    if( cube[11] == 1 )
		    {
		      cube[11] = label;        
		      octreeLabeling( 6, label, cube);
		    }
		    if( cube[13] == 1 )
		    {
		      cube[13] = label;        
		      octreeLabeling( 4, label, cube);
		      octreeLabeling( 6, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		  }
		  if( octant==3 )
		  {
		    if( cube[3] == 1 )
		    {
		      cube[3] = label;        
		      octreeLabeling( 1, label, cube);
		    }
		    if( cube[4] == 1 )
		    {
		      cube[4] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 4, label, cube);
		    }
		    if( cube[12] == 1 )
		    {
		      cube[12] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 5, label, cube);
		      octreeLabeling( 7, label, cube);
		    }
		    if( cube[6] == 1 )
		      cube[6] = label;        
		    if( cube[7] == 1 )
		    {
		      cube[7] = label;        
		      octreeLabeling( 4, label, cube);
		    }
		    if( cube[14] == 1 )
		    {
		      cube[14] = label;        
		      octreeLabeling( 7, label, cube);
		    }
		    if( cube[15] == 1 )
		    {
		      cube[15] = label;        
		      octreeLabeling( 4, label, cube);
		      octreeLabeling( 7, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		  }
		  if( octant==4 )
		  {
		  	if( cube[4] == 1 )
		    {
		      cube[4] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 3, label, cube);
		    }
		  	if( cube[5] == 1 )
		    {
		      cube[5] = label;        
		      octreeLabeling( 2, label, cube);
		    }
		    if( cube[13] == 1 )
		    {
		      cube[13] = label;        
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 6, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		    if( cube[7] == 1 )
		    {
		      cube[7] = label;        
		      octreeLabeling( 3, label, cube);
		    }
		    if( cube[15] == 1 )
		    {
		      cube[15] = label;        
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 7, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		    if( cube[8] == 1 )
		      cube[8] = label;        
		    if( cube[16] == 1 )
		    {
		      cube[16] = label;        
		      octreeLabeling( 8, label, cube);
		    }
		  }
		  if( octant==5 )
		  {
		  	if( cube[9] == 1 )
		    {
		      cube[9] = label;        
		      octreeLabeling( 1, label, cube);
		    }
		    if( cube[10] == 1 )
		    {
		      cube[10] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 6, label, cube);
		    }
		    if( cube[12] == 1 )
		    {
		      cube[12] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 7, label, cube);
		    }
		    if( cube[17] == 1 )
		      cube[17] = label;        
		    if( cube[18] == 1 )
		    {
		      cube[18] = label;        
		      octreeLabeling( 6, label, cube);
		    }
		    if( cube[20] == 1 )
		    {
		      cube[20] = label;        
		      octreeLabeling( 7, label, cube);
		    }
		    if( cube[21] == 1 )
		    {
		      cube[21] = label;        
		      octreeLabeling( 6, label, cube);
		      octreeLabeling( 7, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		  }
		  if( octant==6 )
		  {
		  	if( cube[10] == 1 )
		    {
		      cube[10] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 5, label, cube);
		    }
		    if( cube[11] == 1 )
		    {
		      cube[11] = label;        
		      octreeLabeling( 2, label, cube);
		    }
		    if( cube[13] == 1 )
		    {
		      cube[13] = label;        
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 4, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		    if( cube[18] == 1 )
		    {
		      cube[18] = label;        
		      octreeLabeling( 5, label, cube);
		    }
		    if( cube[21] == 1 )
		    {
		      cube[21] = label;        
		      octreeLabeling( 5, label, cube);
		      octreeLabeling( 7, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		    if( cube[19] == 1 )
		      cube[19] = label;        
		    if( cube[22] == 1 )
		    {
		      cube[22] = label;        
		      octreeLabeling( 8, label, cube);
		    }
		  }
		  if( octant==7 )
		  {
		  	if( cube[12] == 1 )
		    {
		      cube[12] = label;        
		      octreeLabeling( 1, label, cube);
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 5, label, cube);
		    }
		  	if( cube[14] == 1 )
		    {
		      cube[14] = label;        
		      octreeLabeling( 3, label, cube);
		    }
		    if( cube[15] == 1 )
		    {
		      cube[15] = label;        
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 4, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		    if( cube[20] == 1 )
		    {
		      cube[20] = label;        
		      octreeLabeling( 5, label, cube);
		    }
		    if( cube[21] == 1 )
		    {
		      cube[21] = label;        
		      octreeLabeling( 5, label, cube);
		      octreeLabeling( 6, label, cube);
		      octreeLabeling( 8, label, cube);
		    }
		    if( cube[23] == 1 )
		      cube[23] = label;        
		    if( cube[24] == 1 )
		    {
		      cube[24] = label;        
		      octreeLabeling( 8, label, cube);
		    }
		  }
		  if( octant==8 )
		  {
		  	if( cube[13] == 1 )
		    {
		      cube[13] = label;        
		      octreeLabeling( 2, label, cube);
		      octreeLabeling( 4, label, cube);
		      octreeLabeling( 6, label, cube);
		    }
		  	if( cube[15] == 1 )
		    {
		      cube[15] = label;        
		      octreeLabeling( 3, label, cube);
		      octreeLabeling( 4, label, cube);
		      octreeLabeling( 7, label, cube);
		    }
		  	if( cube[16] == 1 )
		    {
		      cube[16] = label;        
		      octreeLabeling( 4, label, cube);
		    }
		  	if( cube[21] == 1 )
		    {
		      cube[21] = label;        
		      octreeLabeling( 5, label, cube);
		      octreeLabeling( 6, label, cube);
		      octreeLabeling( 7, label, cube);
		    }
		  	if( cube[22] == 1 )
		    {
		      cube[22] = label;        
		      octreeLabeling( 6, label, cube);
		    }
		  	if( cube[24] == 1 )
		    {
		      cube[24] = label;        
		      octreeLabeling( 7, label, cube);
		    }
		  	if( cube[25] == 1 )
		      cube[25] = label;        
		  }
}
/**
 * Check if current point is a Simple Point.
 * This method is named 'N(v)_labeling' in [Lee94].
 * Outputs the number of connected objects in a neighborhood of a point
 * after this point would have been removed.
 * 
 * @param neighbors neighbor pixels of the point
 * @return true or false if the point is simple or not
 */
inline bool isSimplePoint(const std::vector<const bool*> &neighbors) 
{
		// copy neighbors for labeling
		int cube[26]; 
		int i;
		for ( i = 0; i < 13; i++ )  // i =  0..12 -> cube[0..12]
			cube[i] = *(neighbors[i]);
		// i != 13 : ignore center pixel when counting (see [Lee94])
		for ( i = 14; i < 27; i++ ) // i = 14..26 -> cube[13..25]
			cube[i-1] = *(neighbors[i]);
		// set initial label
		int label = 2;
		// for all points in the neighborhood
		for( i = 0; i < 26; i++ )
		{
			if( cube[i]==1 )     // voxel has not been labeled yet
			{
				// start recursion with any octant that contains the point i
				switch( i )
				{
				case 0:
				case 1:
				case 3:
				case 4:
				case 9:
				case 10:
				case 12:
					octreeLabeling(1, label, cube );
					break;
				case 2:
				case 5:
				case 11:
				case 13:
					octreeLabeling(2, label, cube );
					break;
				case 6:
				case 7:
				case 14:
				case 15:
					octreeLabeling(3, label, cube );
					break;
				case 8:
				case 16:
					octreeLabeling(4, label, cube );
					break;
				case 17:
				case 18:
				case 20:
				case 21:
					octreeLabeling(5, label, cube );
					break;
				case 19:
				case 22:
					octreeLabeling(6, label, cube );
					break;
				case 23:
				case 24:
					octreeLabeling(7, label, cube );
					break;
				case 25:
					octreeLabeling(8, label, cube );
					break;
				}
				label++;

				if( label-2 >= 2 )
				{
					return false;
				}
			}
		}

		return true;
}

I3D_ALGO_EXPORT
void Skeletonize(const i3d::Image3d<bool> &img, i3d::Image3d<bool> &skeleton) 
{
	   skeleton = img;

	   skeleton.PadImage(1, false);

		int width = skeleton.GetSizeX();
		int height = skeleton.GetSizeY();
		int depth = skeleton.GetSizeZ();
		int sliceSize = skeleton.GetSliceSize();
		std::vector<const bool*> win;

		// Prepare Euler LUT 
		int eulerLUT[256];
		fillEulerLUT( eulerLUT );
		
		std::vector<i3d::Vector3d<int>/**/> simpleBorderPoints;
		
		// Loop through the image several times until there is no change.
		int unchangedBorders = 0;
		while (unchangedBorders < 6)  // loop until no change for all the six border types
		{						
			unchangedBorders = 0;

			for (int currentBorder = 1; currentBorder <= 6; currentBorder++)
			{
				bool noChange = true;				
				
				// Loop through the image.				 
				for (int z = 1; z < depth-1; z++)
				{
					for (int y = 1; y < height-1; y++)
					{
						bool *imgPtr = skeleton.GetVoxelAddr(1,y,z);
						for (int x = 1; x < width-1; x++)						
						{
							// check if point is foreground
							if ( *imgPtr != true) 
							{
								 imgPtr++;
								 continue;         // current point is already background 
							}
																				
							// check 6-neighbors if point is a border point of type currentBorder
							bool isBorderPoint = false;
							
							// North
							if ((currentBorder == 1) && (*(imgPtr-width) == false))
								isBorderPoint = true;

							// South
							if ((currentBorder == 2) && (*(imgPtr+width) == false))
								isBorderPoint = true;

							// East
							if ((currentBorder == 3) && (*(imgPtr+1) == false))
								isBorderPoint = true;

							// West
							if ((currentBorder == 4) && (*(imgPtr-1) == false))
								isBorderPoint = true;

							// Up							
							if ((currentBorder == 5) && (*(imgPtr+sliceSize) == false))
								isBorderPoint = true;

							// Bottom
							if ((currentBorder == 6) && (*(imgPtr-sliceSize) == false))
								isBorderPoint = true;
								

							if( !isBorderPoint )
							{
								 imgPtr++;
								 continue;         // current point is not deletable
							}

							i3d::MakeWindow(skeleton, x,y,z, nb3D_full, win); 
							
 							// Check if a point in the given stack is at the end of an arc
							int numberOfNeighbors = -1;  
							for ( int i = 0; i < count; i++ ) 
							{					        	
								 numberOfNeighbors += *(win[i]);
							}

							if ( numberOfNeighbors == 1 ) // is end point, i.e., not deletable
							{
								 imgPtr++;
								 continue;
							}

							// Check if point is Euler invariant (condition 1 in Lee[94])
							if( !isEulerInvariant( win, eulerLUT ) )
							{
								 imgPtr++;
								 continue;         // current point is not deletable
							}

							// Check if point is simple 
							// (deletion does not change connectivity in the 3x3x3 neighborhood)
							// (conditions 2 and 3 in Lee[94])
							if( !isSimplePoint(win) )
							{
								 imgPtr++;
								 continue;         // current point is not deletable
							}

							// add all simple border points to a list for sequential re-checking
							simpleBorderPoints.push_back(i3d::Vector3d<int>(x,y,z));

							imgPtr++;
						}
					}					
				}							


				// sequential re-checking to preserve connectivity when
				// deleting in a parallel way
				for (size_t i=0; i<simpleBorderPoints.size(); i++)
				{
					 i3d::Vector3d<int> point = simpleBorderPoints[i];
					 i3d::MakeWindow(skeleton, point.x, point.y, point.z, nb3D_full, win);

					// Check if border points is simple			        
					if (isSimplePoint(win)) 
					{
						 // we can delete the current point
						 skeleton.SetVoxel(point, false);
						 noChange = false;
					}


				}

				if (noChange)
				{
					unchangedBorders++;
				}

				simpleBorderPoints.clear();
			} 
		}

		skeleton.RemovePadding(1);
} 

} // namespace i3d
	
