/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file watersheds_lin.cc
  \brief Part of mathematical morphology routines - implementation.

  This file contains bodies of mathematical morphology routines dealing with 
  the watershed transformation algorithms by Lin.
*/ 

#include "watersheds.h"
#include "i3dassert.h"
#include "ivLab/toboggan.h"
#include "ivLab/watershed.h"
#include "ivLab3D/toboggan.h"
#include "ivLab3D/watershed.h"


#ifdef __GNUG__
#	pragma implementation
#endif

namespace i3d {

	//--------------------------------------------------------------------------
	// Watershed tobogan implementation - uses code from ivLab3D
	//--------------------------------------------------------------------------

	template <class T> void WatershedToboggan(const Image3d<T> &in_grad, Image3d<int> &out, int connectivity, bool drawLines)
	{
		I3DASSERT_ALWAYS_ET(connectivity == 4 || connectivity == 8 || connectivity == 6 || connectivity == 18 || connectivity == 26,
			"Unsupported connectivity required");

		out.MakeRoom(in_grad.GetSize());

		int *result = ivLab3D::Toboggan(in_grad.GetFirstVoxelAddr(), in_grad.GetWidth(), in_grad.GetHeight(), in_grad.GetNumSlices(), true, connectivity, drawLines);
		if (result) {
			memcpy(out.GetFirstVoxelAddr(), result, out.GetImageSize()*sizeof(int));
			delete [] result;
			result = NULL;
		}
	}

	template I3D_ALGO_EXPORT void WatershedToboggan(const Image3d<i3d::GRAY8> &in_grad, Image3d<int> &out, int connectivity, bool drawLines);
	template I3D_ALGO_EXPORT void WatershedToboggan(const Image3d<i3d::GRAY16> &in_grad, Image3d<int> &out, int connectivity, bool drawLines);
	template I3D_ALGO_EXPORT void WatershedToboggan(const Image3d<float> &in_grad, Image3d<int> &out, int connectivity, bool drawLines);

	//--------------------------------------------------------------------------
	// Watershed tobogan implementation - uses code from ivLab
	//--------------------------------------------------------------------------

	template <class T> void WatershedToboggan2D(const Image3d<T> &in_grad, Image3d<int> &out, int connectivity)
	{
		I3DASSERT_ALWAYS_ET(connectivity == 4 || connectivity == 8, "Unsupported connectivity required");
		I3DASSERT_ALWAYS_ET(in_grad.GetSizeZ() == 1, "Algorithm works for 2D images only");

		out.MakeRoom(in_grad.GetSize());

		int *result = ivLab::Toboggan(in_grad.GetFirstVoxelAddr(), in_grad.GetWidth(), in_grad.GetHeight(), true, connectivity);
		if (result) {
			memcpy(out.GetFirstVoxelAddr(), result, out.GetImageSize()*sizeof(int));
			delete [] result;
			result = NULL;
		}
	}

	template I3D_ALGO_EXPORT void WatershedToboggan2D(const Image3d<i3d::GRAY8> &in_grad, Image3d<int> &out, int connectivity);
	template I3D_ALGO_EXPORT void WatershedToboggan2D(const Image3d<i3d::GRAY16> &in_grad, Image3d<int> &out, int connectivity);
	template I3D_ALGO_EXPORT void WatershedToboggan2D(const Image3d<float> &in_grad, Image3d<int> &out, int connectivity);

	//--------------------------------------------------------------------------
	// Watershed immersion implementation - uses code from ivLab3D
	//--------------------------------------------------------------------------

	template <class T> void WatershedLin(const Image3d<T> &in_grad, Image3d<int> &out, int connectivity, bool drawLines)
	{
		I3DASSERT_ALWAYS_ET(connectivity == 4 || connectivity == 8 || connectivity == 6 || connectivity == 18 || connectivity == 26,
			"Unsupported connectivity required");

		out.MakeRoom(in_grad.GetSize());

		int *result = ivLab3D::Watershed(in_grad.GetFirstVoxelAddr(), in_grad.GetWidth(), in_grad.GetHeight(), in_grad.GetNumSlices(), true, connectivity, drawLines);
		if (result) {
			memcpy(out.GetFirstVoxelAddr(), result, out.GetImageSize()*sizeof(int));
			delete [] result;
			result = NULL;
		}
	}

	template I3D_ALGO_EXPORT void WatershedLin(const Image3d<i3d::GRAY8> &in_grad, Image3d<int> &out, int connectivity, bool drawLines);
	template I3D_ALGO_EXPORT void WatershedLin(const Image3d<i3d::GRAY16> &in_grad, Image3d<int> &out, int connectivity, bool drawLines);
	template I3D_ALGO_EXPORT void WatershedLin(const Image3d<float> &in_grad, Image3d<int> &out, int connectivity, bool drawLines);

	//--------------------------------------------------------------------------
	// Watershed immersion implementation - uses code from ivLab
	//--------------------------------------------------------------------------

	template <class T> void WatershedLin2D(const Image3d<T> &in_grad, Image3d<int> &out, int connectivity)
	{
		I3DASSERT_ALWAYS_ET(connectivity == 4 || connectivity == 8, "Unsupported connectivity required");
		I3DASSERT_ALWAYS_ET(in_grad.GetSizeZ() == 1, "Algorithm works for 2D images only");

		out.MakeRoom(in_grad.GetSize());

		int *result = ivLab::Watershed(in_grad.GetFirstVoxelAddr(), in_grad.GetWidth(), in_grad.GetHeight(), true, connectivity);
		if (result) {
			memcpy(out.GetFirstVoxelAddr(), result, out.GetImageSize()*sizeof(int));
			delete [] result;
			result = NULL;
		}
	}

	template I3D_ALGO_EXPORT void WatershedLin2D(const Image3d<i3d::GRAY8> &in_grad, Image3d<int> &out, int connectivity);
	template I3D_ALGO_EXPORT void WatershedLin2D(const Image3d<i3d::GRAY16> &in_grad, Image3d<int> &out, int connectivity);
	template I3D_ALGO_EXPORT void WatershedLin2D(const Image3d<float> &in_grad, Image3d<int> &out, int connectivity);

} // namespace i3d
